/*
Copyright 2021 The KodeRover Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package httpclient

import (
	"bytes"
	"io"
	"os"

	"github.com/go-resty/resty/v2"
)

const UserAgent = "Zadig REST Client"

type Client struct {
	*resty.Client

	Host    string // http://example.org
	BaseURI string // /api/v1
}

func Get(url string, rfs ...RequestFunc) (*resty.Response, error) {
	return New().Get(url, rfs...)
}

func Post(url string, rfs ...RequestFunc) (*resty.Response, error) {
	return New().Post(url, rfs...)
}

func Patch(url string, rfs ...RequestFunc) (*resty.Response, error) {
	return New().Patch(url, rfs...)
}

func Put(url string, rfs ...RequestFunc) (*resty.Response, error) {
	return New().Put(url, rfs...)
}

func Delete(url string, rfs ...RequestFunc) (*resty.Response, error) {
	return New().Delete(url, rfs...)
}

func Head(url string, rfs ...RequestFunc) (*resty.Response, error) {
	return New().Head(url, rfs...)
}

func Options(url string, rfs ...RequestFunc) (*resty.Response, error) {
	return New().Options(url, rfs...)
}

// Download retrieves content from the given url and write it to path.
func Download(url, path string, rfs ...RequestFunc) error {
	res, err := Get(url, rfs...)
	if err != nil {
		return err
	}

	f, err := os.Create(path)
	if err != nil {
		return err
	}
	defer func() {
		_ = f.Close()
	}()

	_, err = io.Copy(f, bytes.NewReader(res.Body()))

	return err
}

func New(cfs ...ClientFunc) *Client {
	r := resty.New()
	r.SetHeader("Content-Type", "application/json").
		SetHeader("Accept", "application/json").
		SetHeader("User-Agent", UserAgent)

	c := &Client{
		Client: r,
	}

	for _, cf := range cfs {
		cf(c)
	}

	return c
}

func (c *Client) Get(url string, rfs ...RequestFunc) (*resty.Response, error) {
	return c.Request(resty.MethodGet, url, rfs...)
}

func (c *Client) Post(url string, rfs ...RequestFunc) (*resty.Response, error) {
	return c.Request(resty.MethodPost, url, rfs...)
}

func (c *Client) Patch(url string, rfs ...RequestFunc) (*resty.Response, error) {
	return c.Request(resty.MethodPatch, url, rfs...)
}

func (c *Client) Put(url string, rfs ...RequestFunc) (*resty.Response, error) {
	return c.Request(resty.MethodPut, url, rfs...)
}

func (c *Client) Delete(url string, rfs ...RequestFunc) (*resty.Response, error) {
	return c.Request(resty.MethodDelete, url, rfs...)
}

func (c *Client) Head(url string, rfs ...RequestFunc) (*resty.Response, error) {
	return c.Request(resty.MethodHead, url, rfs...)
}

func (c *Client) Options(url string, rfs ...RequestFunc) (*resty.Response, error) {
	return c.Request(resty.MethodOptions, url, rfs...)
}

func (c *Client) Request(method, url string, rfs ...RequestFunc) (*resty.Response, error) {
	if c.BaseURI != "" {
		url = c.BaseURI + url
	}
	r := c.R()

	for _, rf := range rfs {
		rf(r)
	}

	return wrapError(r.Execute(method, url))
}

func wrapError(res *resty.Response, err error) (*resty.Response, error) {
	if err != nil {
		return nil, err
	}

	if res.IsError() {
		return nil, &Error{Code: res.StatusCode(), Status: res.Status(), Detail: res.Body()}
	}

	return res, nil
}
